//---------------------------------------------------------------------------
// TMySkin V1.0 par BRUN Stephane 07/2003
// <brun.stephane@free.fr>
//
// Composant graphique driv de TImage
// prend la taille de la form et se sert de l'image pour skinner (dcouper)
// la fentre de l'application.
//
// Propriets :
//  Tcolor TransparentColor : Couleur qui dfinit la transparance
//  Bool SkinMove : authorise ou non le dplacement de la fentre en cliquant
//                  sur l'image. (Pas encore implement)
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "MySkin.h"     
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(TMySkin *)
{
        new TMySkin(NULL);
}
//---------------------------------------------------------------------------
__fastcall TMySkin::TMySkin(TComponent* Owner)
        : TImage(Owner)
{
    // Init ICI
    hRgn=NULL;
    FTransparence=255;

    Align = alClient;
    AutoSize = false;
    Stretch = false;
    Picture->OnChange = SkinUpdate;
}
//---------------------------------------------------------------------------
__fastcall TMySkin::~TMySkin()
{
    if(hRgn)
    {
        DeleteObject(hRgn);
    }
    hRgn = NULL;
}

//---------------------------------------------------------------------------
namespace Myskin
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TMySkin)};
        RegisterComponents("MyComp", classes, 0);
    }
}

//---------------------------------------------------------------------------
void __fastcall TMySkin::EnableRegion(void)
{
    GetRegion();
    Visible = false;
    DrawRegion();
    Visible = true;
}

//---------------------------------------------------------------------------
void __fastcall TMySkin::DisableRegion(void)
{
    SetWindowRgn(Parent->Handle, NULL, false);
    hRgn = NULL;
}

//---------------------------------------------------------------------------
void __fastcall TMySkin::DrawRegion(void)
{
    if(!ComponentState.Contains(csDesigning) && hRgn)
    {
	    SetWindowRgn(Parent->Handle, hRgn, true);
    }
}

//---------------------------------------------------------------------------
void __fastcall TMySkin::GetRegion(void)
{
//    ShowMessage(IntToHex(FTransparentColor,8));

    BYTE cTolerance = 10;
    HBITMAP hBmp = Picture->Bitmap->Handle;

    // Pas de rgion lors du design
    if(ComponentState.Contains(csDesigning))
        return;
    if(!Parent)
        return;
    if(Picture->Bitmap->Empty)
        return;

    hRgn = NULL;

    if(hBmp)
	{
		// Create a memory DC inside which we will scan the bitmap content
		HDC hMemDC = CreateCompatibleDC(NULL);
		if (hMemDC)
		{
			// Get bitmap size
			BITMAP bm;
			GetObject(hBmp, sizeof(bm), &bm);

			// Create a 32 bits depth bitmap and select it into the memory DC
			BITMAPINFOHEADER RGB32BITSBITMAPINFO = {
					sizeof(BITMAPINFOHEADER),	// biSize
					bm.bmWidth,					// biWidth;
					bm.bmHeight,				// biHeight;
					1,							// biPlanes;
					32,							// biBitCount
					BI_RGB,						// biCompression;
					0,							// biSizeImage;
					0,							// biXPelsPerMeter;
					0,							// biYPelsPerMeter;
					0,							// biClrUsed;
					0							// biClrImportant;
			};
			VOID * pbits32;
			HBITMAP hbm32 = CreateDIBSection(hMemDC, (BITMAPINFO *)&RGB32BITSBITMAPINFO, DIB_RGB_COLORS, &pbits32, NULL, 0);
			if (hbm32)
			{
				HBITMAP holdBmp = (HBITMAP)SelectObject(hMemDC, hbm32);

				// Create a DC just to copy the bitmap into the memory DC
				HDC hDC = CreateCompatibleDC(hMemDC);
				if (hDC)
				{
					// Get how many bytes per row we have for the bitmap bits (rounded up to 32 bits)
					BITMAP bm32;
					GetObject(hbm32, sizeof(bm32), &bm32);
					while (bm32.bmWidthBytes % 4)
						bm32.bmWidthBytes++;

					// Copy the bitmap into the memory DC
					HBITMAP holdBmp = (HBITMAP)SelectObject(hDC, hBmp);
					BitBlt(hMemDC, 0, 0, bm.bmWidth, bm.bmHeight, hDC, 0, 0, SRCCOPY);

					// For better performances, we will use the ExtCreateRegion() function to create the
					// region. This function take a RGNDATA structure on entry. We will add rectangles by
					// amount of ALLOC_UNIT number in this structure.
					#define ALLOC_UNIT	100
					DWORD maxRects = ALLOC_UNIT;
					HANDLE hData = GlobalAlloc(GMEM_MOVEABLE, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects));
					RGNDATA *pData = (RGNDATA *)GlobalLock(hData);
					pData->rdh.dwSize = sizeof(RGNDATAHEADER);
					pData->rdh.iType = RDH_RECTANGLES;
					pData->rdh.nCount = pData->rdh.nRgnSize = 0;
					SetRect(&pData->rdh.rcBound, MAXLONG, MAXLONG, 0, 0);

					// Keep on hand highest and lowest values for the "transparent" pixels
					BYTE lr = GetRValue(FTransparentColor);
					BYTE lg = GetGValue(FTransparentColor);
					BYTE lb = GetBValue(FTransparentColor);

                    BYTE hr = lr;
                    BYTE hg = lg;
                    BYTE hb = lb;
                    if(cTolerance!=0)
                    {
                      hr = min(0xff, lr + cTolerance);//GetRValue(cTolerance));
                      hg = min(0xff, lg + cTolerance);//GetGValue(cTolerance));
                      hb = min(0xff, lb + cTolerance);//GetBValue(cTolerance));
                      lr = max(0x00, lr - cTolerance);//GetRValue(cTolerance));
                      lg = max(0x00, lg - cTolerance);//GetGValue(cTolerance));
                      lb = max(0x00, lb - cTolerance);//GetBValue(cTolerance));
                    }

					// Scan each bitmap row from bottom to top (the bitmap is inverted vertically)
					BYTE *p32 = (BYTE *)bm32.bmBits + (bm32.bmHeight - 1) * bm32.bmWidthBytes;
					for (int y = 0; y < bm.bmHeight; y++)
					{
						// Scan each bitmap pixel from left to right
						for (int x = 0; x < bm.bmWidth; x++)
						{
							// Search for a continuous range of "non transparent pixels"
							int x0 = x;
							LONG *p = (LONG *)p32 + x;
							while (x < bm.bmWidth)
							{
								BYTE b = GetRValue(*p);
								if (b >= lr && b <= hr)
								{
									b = GetGValue(*p);
									if (b >= lg && b <= hg)
									{
										b = GetBValue(*p);
										if (b >= lb && b <= hb)
											// This pixel is "transparent"
											break;
									}
								}
								p++;
								x++;
							}

							if (x > x0)
							{
								// Add the pixels (x0, y) to (x, y+1) as a new rectangle in the region
								if (pData->rdh.nCount >= maxRects)
								{
									GlobalUnlock(hData);
									maxRects += ALLOC_UNIT;
									hData = GlobalReAlloc(hData, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), GMEM_MOVEABLE);
									pData = (RGNDATA *)GlobalLock(hData);
								}
								RECT *pr = (RECT *)&pData->Buffer;
								SetRect(&pr[pData->rdh.nCount], x0, y, x, y+1);
								if (x0 < pData->rdh.rcBound.left)
									pData->rdh.rcBound.left = x0;
								if (y < pData->rdh.rcBound.top)
									pData->rdh.rcBound.top = y;
								if (x > pData->rdh.rcBound.right)
									pData->rdh.rcBound.right = x;
								if (y+1 > pData->rdh.rcBound.bottom)
									pData->rdh.rcBound.bottom = y+1;
								pData->rdh.nCount++;

								// On Windows98, ExtCreateRegion() may fail if the number of rectangles is too
								// large (ie: > 4000). Therefore, we have to create the region by multiple steps.
								if (pData->rdh.nCount == 2000)
								{
									HRGN h = ExtCreateRegion(NULL, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), pData);
									if (hRgn)
									{
										CombineRgn(hRgn, hRgn, h, RGN_OR);
										DeleteObject(h);
									}
									else
										hRgn = h;
									pData->rdh.nCount = 0;
									SetRect(&pData->rdh.rcBound, MAXLONG, MAXLONG, 0, 0);
								}
							}
						}

						// Go to next row (remember, the bitmap is inverted vertically)
						p32 -= bm32.bmWidthBytes;
					}

					// Create or extend the region with the remaining rectangles
					HRGN h = ExtCreateRegion(NULL, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), pData);
					if (hRgn)
					{
						CombineRgn(hRgn, hRgn, h, RGN_OR);
						DeleteObject(h);
					}
					else
						hRgn = h;

					// Clean up
					GlobalFree(hData);
					SelectObject(hDC, holdBmp);
					DeleteDC(hDC);
				} // fin if hDC

				DeleteObject(SelectObject(hMemDC, holdBmp));
			} // fin if hbm32
			DeleteDC(hMemDC);
		} // Fin if hMemDC
	}// fin if hBmp

    // Thanks to Boian Mitov for the OffsetRgn code
    if(hRgn)
    {
        TPoint ClientPos = Parent->ClientToScreen(::Point(0, 0));

        if(Parent->Parent)
        {
            TPoint ParentPos = Parent->Parent->ClientToScreen(::Point(Parent->Left, Parent->Top));
            ClientPos.x -= ParentPos.x;
            ClientPos.y -= ParentPos.y;
        }
        else
        {
            ClientPos.x -= Parent->Left;
            ClientPos.y -= Parent->Top;
        }
        OffsetRgn(hRgn, ClientPos.x, ClientPos.y);
    }
}

void __fastcall TMySkin::SkinUpdate(System::TObject* /*Sender*/)
{
    GetRegion();
    Visible = false;
    DrawRegion();
    SetTransparence(FTransparence);
    Visible = true;
}

void __fastcall TMySkin::Loaded(void)
{
    inherited::Loaded();

    GetRegion();
    Visible = false;
    DrawRegion();
    SetTransparence(FTransparence);
    Visible = true;
}

//---------------------------------------------------------------------------
// Thanks to Boian Mitov for the following code
//---------------------------------------------------------------------------
void __fastcall TMySkin::SetParent(Controls::TWinControl* Value)
{
    inherited::SetParent(Value);

    if(Value)
    {
        if(dynamic_cast <TForm *> (Value))
        {
            for(int i=0; i<Value->ControlCount; i++)
            {
                if(dynamic_cast <TMySkin *>(Value->Controls[i]) && Value->Controls[i] != this)
                    throw new Exception("Please Drop only one TBMShapedForm per Form");
            }
        }
        else
            throw new Exception("Please Drop on a Form");
    }
}
//---------------------------------------------------------------------------
// End of Boian Mitov code
//---------------------------------------------------------------------------

void __fastcall TMySkin::SetTransparence(int t)
{
    HWND window_handle = Parent->Handle;

    if(t<0)
        t=0;
    else if(t>=255)
        t=255;

    FTransparence=t;

    if(ComponentState.Contains(csDesigning))
        return;

    OSVERSIONINFO OSVersion;
    OSVersion.dwOSVersionInfoSize=sizeof(OSVERSIONINFO);
    GetVersionEx(&OSVersion);

    // Si on est sous Windows 2000 ou XP ou +
    if( (OSVersion.dwMajorVersion>=5) && (OSVersion.dwPlatformId == VER_PLATFORM_WIN32_NT) )
    {
        SetWindowLong(window_handle, GWL_EXSTYLE, GetWindowLong(window_handle, GWL_EXSTYLE) | WS_EX_LAYERED);
        SetLayeredWindowAttributes(window_handle, 0, FTransparence, LWA_ALPHA);
    }
}

/*
MESSAGE void __fastcall TMySkin::MoveWin(TMessage &Msg)
{
    TMySkin::Dispatch(&Msg);

    if(ComponentState.Contains(csDesigning))
        return;

    if((FSkinMove==true) && (Msg.Result == HTCLIENT))
    {
        Msg.Result = HTCAPTION;
    }
}
*/

